@@ -4,8 +4,6 @@
 
 #include "daemon/bindings/impl.h"
 
-#include "daemon/zimport.h"
-
 /** @internal return cache, or throw lua error if not open */
 static struct kr_cache * cache_assert_open(lua_State *L)
 {
@@ -130,10 +128,10 @@ static int cache_max_ttl(lua_State *L)
 			lua_error_p(L, "expected 'max_ttl(number ttl)'");
 		uint32_t min = cache->ttl_min;
 		int64_t ttl = lua_tointeger(L, 1);
-		if (ttl < 1 || ttl < min || ttl > UINT32_MAX) {
+		if (ttl < 1 || ttl < min || ttl > TTL_MAX_MAX) {
 			lua_error_p(L,
 				"max_ttl must be larger than minimum TTL, and in range <1, "
-				STR(UINT32_MAX) ">'");
+				STR(TTL_MAX_MAX) ">'");
 		}
 		cache->ttl_max = ttl;
 	}
@@ -152,10 +150,10 @@ static int cache_min_ttl(lua_State *L)
 			lua_error_p(L, "expected 'min_ttl(number ttl)'");
 		uint32_t max = cache->ttl_max;
 		int64_t ttl = lua_tointeger(L, 1);
-		if (ttl < 0 || ttl > max || ttl > UINT32_MAX) {
+		if (ttl < 0 || ttl > max || ttl > TTL_MAX_MAX) {
 			lua_error_p(L,
 				"min_ttl must be smaller than maximum TTL, and in range <0, "
-				STR(UINT32_MAX) ">'");
+				STR(TTL_MAX_MAX) ">'");
 		}
 		cache->ttl_min = ttl;
 	}
@@ -361,81 +359,6 @@ static int cache_ns_tout(lua_State *L)
 	return 1;
 }
 
-/** Zone import completion callback.
- * Deallocates zone import context. */
-static void cache_zone_import_cb(int state, void *param)
-{
-	(void)state;
-	struct worker_ctx *worker = param;
-	if (kr_fails_assert(worker && worker->z_import)) return;
-	zi_free(worker->z_import);
-	worker->z_import = NULL;
-}
-
-/** Import zone from file. */
-static int cache_zone_import(lua_State *L)
-{
-	int ret = -1;
-	char msg[128];
-
-	struct worker_ctx *worker = the_worker;
-	if (!worker) {
-		strncpy(msg, "internal error, empty worker pointer", sizeof(msg));
-		goto finish;
-	}
-
-	if (worker->z_import && zi_import_started(worker->z_import)) {
-		strncpy(msg, "import already started", sizeof(msg));
-		goto finish;
-	}
-
-	(void)cache_assert_open(L); /* just check it in advance */
-
-	/* Check parameters */
-	int n = lua_gettop(L);
-	if (n < 1 || !lua_isstring(L, 1)) {
-		strncpy(msg, "expected 'cache.zone_import(path to zone file)'", sizeof(msg));
-		goto finish;
-	}
-
-	/* Parse zone file */
-	const char *zone_file = lua_tostring(L, 1);
-
-	const char *default_origin = NULL; /* TODO */
-	uint16_t default_rclass = 1;
-	uint32_t default_ttl = 0;
-
-	if (worker->z_import == NULL) {
-		worker->z_import = zi_allocate(worker, cache_zone_import_cb, worker);
-		if (worker->z_import == NULL) {
-			strncpy(msg, "can't allocate zone import context", sizeof(msg));
-			goto finish;
-		}
-	}
-
-	ret = zi_zone_import(worker->z_import, zone_file, default_origin,
-			     default_rclass, default_ttl);
-
-	lua_newtable(L);
-	if (ret == 0) {
-		strncpy(msg, "zone file successfully parsed, import started", sizeof(msg));
-	} else if (ret == 1) {
-		strncpy(msg, "TA not found", sizeof(msg));
-	} else {
-		strncpy(msg, "error parsing zone file", sizeof(msg));
-	}
-
-finish:
-	msg[sizeof(msg) - 1] = 0;
-	lua_newtable(L);
-	lua_pushstring(L, msg);
-	lua_setfield(L, -2, "msg");
-	lua_pushnumber(L, ret);
-	lua_setfield(L, -2, "code");
-
-	return 1;
-}
-
 int kr_bindings_cache(lua_State *L)
 {
 	static const luaL_Reg lib[] = {
@@ -450,7 +373,6 @@ int kr_bindings_cache(lua_State *L)
 		{ "max_ttl", cache_max_ttl },
 		{ "min_ttl", cache_min_ttl },
 		{ "ns_tout", cache_ns_tout },
-		{ "zone_import", cache_zone_import },
 		{ NULL, NULL }
 	};
 
