@@ -6,11 +6,11 @@
 
 #include "lib/defines.h"
 #include "daemon/session.h"
-#include "daemon/engine.h"
 #include "daemon/tls.h"
 #include "daemon/http.h"
 #include "daemon/worker.h"
 #include "daemon/io.h"
+#include "daemon/proxyv2.h"
 #include "lib/generic/queue.h"
 
 #define TLS_CHUNK_SIZE (16 * 1024)
@@ -28,14 +28,16 @@
  */
 struct session {
 	struct session_flags sflags;  /**< miscellaneous flags. */
-	union inaddr peer;            /**< address of peer; not for UDP clients (downstream) */
-	union inaddr sockname;        /**< our local address; for UDP it may be a wildcard */
+	union kr_sockaddr peer;       /**< address of peer; not for UDP clients (downstream) */
+	union kr_sockaddr sockname;   /**< our local address; for UDP it may be a wildcard */
 	uv_handle_t *handle;          /**< libuv handle for IO operations. */
 	uv_timer_t timeout;           /**< libuv handle for timer. */
 
 	struct tls_ctx *tls_ctx;      /**< server side tls-related data. */
 	struct tls_client_ctx *tls_client_ctx;  /**< client side tls-related data. */
 
+	struct proxy_result *proxy;   /**< PROXYv2 data for TCP. May be `NULL` if not proxied. */
+
 #if ENABLE_DOH2
 	struct http_ctx *http_ctx;  /**< server side http-related data. */
 #endif
@@ -83,6 +85,7 @@ void session_clear(struct session *session)
 	if (session->handle && session->handle->type == UV_TCP) {
 		free(session->wire_buf);
 	}
+	free(session->proxy);
 #if ENABLE_DOH2
 	http_free(session->http_ctx);
 #endif
@@ -108,7 +111,7 @@ void session_close(struct session *session)
 	if (!uv_is_closing((uv_handle_t *)&session->timeout)) {
 		uv_timer_stop(&session->timeout);
 		if (session->tls_client_ctx) {
-			tls_close(&session->tls_client_ctx->c);
+			tls_client_close(session->tls_client_ctx);
 		}
 		if (session->tls_ctx) {
 			tls_close(&session->tls_ctx->c);
@@ -436,6 +439,21 @@ void session_waitinglist_finalize(struct session *session, int status)
 	}
 }
 
+struct proxy_result *session_proxy_create(struct session *session)
+{
+	if (!kr_fails_assert(!session->proxy)) {
+		session->proxy = calloc(1, sizeof(struct proxy_result));
+		kr_require(session->proxy);
+	}
+
+	return session->proxy;
+}
+
+struct proxy_result *session_proxy_get(struct session *session)
+{
+	return session->proxy;
+}
+
 void session_tasklist_finalize(struct session *session, int status)
 {
 	while (session_tasklist_get_len(session) > 0) {
@@ -458,6 +476,9 @@ int session_tasklist_finalize_expired(struct session *session)
 		trie_val_t *v = trie_it_val(it);
 		struct qr_task *task = (struct qr_task *)*v;
 		if ((now - worker_task_creation_time(task)) >= KR_RESOLVE_TIME_LIMIT) {
+			struct kr_request *req = worker_task_request(task);
+			if (!kr_fails_assert(req))
+				kr_query_inform_timeout(req, req->current_query);
 			queue_push(q, task);
 			worker_task_ref(task);
 		}
@@ -531,25 +552,30 @@ int session_timer_stop(struct session *session)
 
 ssize_t session_wirebuf_consume(struct session *session, const uint8_t *data, ssize_t len)
 {
-	if (data != &session->wire_buf[session->wire_buf_end_idx]) {
-		/* shouldn't happen */
+	if (kr_fails_assert(data == &session->wire_buf[session->wire_buf_end_idx]))
 		return kr_error(EINVAL);
-	}
-
-	if (len < 0) {
-		/* shouldn't happen */
+	if (kr_fails_assert(len >= 0))
 		return kr_error(EINVAL);
-	}
-
-	if (session->wire_buf_end_idx + len > session->wire_buf_size) {
-		/* shouldn't happen */
+	if (kr_fails_assert(session->wire_buf_end_idx + len <= session->wire_buf_size))
 		return kr_error(EINVAL);
-	}
 
 	session->wire_buf_end_idx += len;
 	return len;
 }
 
+ssize_t session_wirebuf_trim(struct session *session, ssize_t len)
+{
+	if (kr_fails_assert(len >= 0))
+		return kr_error(EINVAL);
+	if (kr_fails_assert(session->wire_buf_start_idx + len <= session->wire_buf_size))
+		return kr_error(EINVAL);
+
+	session->wire_buf_start_idx += len;
+	if (session->wire_buf_start_idx > session->wire_buf_end_idx)
+		session->wire_buf_end_idx = session->wire_buf_start_idx;
+	return len;
+}
+
 knot_pkt_t *session_produce_packet(struct session *session, knot_mm_t *mm)
 {
 	session->sflags.wirebuf_error = false;
@@ -741,7 +767,7 @@ void session_unpoison(struct session *session)
 	kr_asan_unpoison(session, sizeof(*session));
 }
 
-int session_wirebuf_process(struct session *session, const struct sockaddr *peer)
+int session_wirebuf_process(struct session *session, struct io_comm_data *comm)
 {
 	int ret = 0;
 	if (session->wire_buf_start_idx == session->wire_buf_end_idx)
@@ -756,7 +782,7 @@ int session_wirebuf_process(struct session *session, const struct sockaddr *peer
 	       (ret < max_iterations)) {
 		if (kr_fails_assert(!session_wirebuf_error(session)))
 			return -1;
-		int res = worker_submit(session, peer, NULL, NULL, NULL, pkt);
+		int res = worker_submit(session, comm, NULL, NULL, pkt);
 		/* Errors from worker_submit() are intentionally *not* handled in order to
 		 * ensure the entire wire buffer is processed. */
 		if (res == kr_ok())
