@@ -19,7 +19,6 @@
 
 #include "contrib/ucw/lib.h"
 #include "contrib/base64.h"
-#include "daemon/io.h"
 #include "daemon/tls.h"
 #include "daemon/worker.h"
 #include "daemon/session.h"
@@ -332,7 +331,12 @@ struct tls_ctx *tls_new(struct worker_ctx *worker)
 		return NULL;
 	}
 
-	int err = gnutls_init(&tls->c.tls_session, GNUTLS_SERVER | GNUTLS_NONBLOCK);
+	int flags = GNUTLS_SERVER | GNUTLS_NONBLOCK;
+#if GNUTLS_VERSION_NUMBER >= 0x030705
+	if (gnutls_check_version("3.7.5"))
+		flags |= GNUTLS_NO_TICKETS_TLS12;
+#endif
+	int err = gnutls_init(&tls->c.tls_session, flags);
 	if (err != GNUTLS_E_SUCCESS) {
 		kr_log_error(TLS, "gnutls_init(): %s (%d)\n", gnutls_strerror_name(err), err);
 		tls_free(tls);
@@ -380,6 +384,17 @@ void tls_close(struct tls_common_ctx *ctx)
 	}
 }
 
+void tls_client_close(struct tls_client_ctx *ctx)
+{
+	/* Store the current session data for potential resumption of this session */
+	if (ctx->params) {
+		gnutls_free(ctx->params->session_data.data);
+		gnutls_session_get_data2(ctx->c.tls_session, &ctx->params->session_data);
+	}
+
+	tls_close(&ctx->c);
+}
+
 void tls_free(struct tls_ctx *tls)
 {
 	if (!tls) {
@@ -842,7 +857,7 @@ tls_client_param_t * tls_client_param_new()
  * \param len[out] output length
  * \param key[out] output buffer
  */
-static bool construct_key(const union inaddr *addr, uint32_t *len, char *key)
+static bool construct_key(const union kr_sockaddr *addr, uint32_t *len, char *key)
 {
 	switch (addr->ip.sa_family) {
 	case AF_INET:
@@ -875,7 +890,7 @@ tls_client_param_t ** tls_client_param_getptr(tls_client_params_t **params,
 			return NULL;
 	}
 	/* Construct the key. */
-	const union inaddr *ia = (const union inaddr *)addr;
+	const union kr_sockaddr *ia = (const union kr_sockaddr *)addr;
 	char key[sizeof(ia->ip6.sin6_port) + sizeof(ia->ip6.sin6_addr)];
 	uint32_t len;
 	if (!construct_key(ia, &len, key))
@@ -887,7 +902,7 @@ tls_client_param_t ** tls_client_param_getptr(tls_client_params_t **params,
 
 int tls_client_param_remove(tls_client_params_t *params, const struct sockaddr *addr)
 {
-	const union inaddr *ia = (const union inaddr *)addr;
+	const union kr_sockaddr *ia = (const union kr_sockaddr *)addr;
 	char key[sizeof(ia->ip6.sin6_port) + sizeof(ia->ip6.sin6_addr)];
 	uint32_t len;
 	if (!construct_key(ia, &len, key))
@@ -1056,6 +1071,10 @@ struct tls_client_ctx *tls_client_ctx_new(tls_client_param_t *entry,
 			     | GNUTLS_ENABLE_FALSE_START
 #endif
 	;
+#if GNUTLS_VERSION_NUMBER >= 0x030705
+	if (gnutls_check_version("3.7.5"))
+		flags |= GNUTLS_NO_TICKETS_TLS12;
+#endif
 	int ret = gnutls_init(&ctx->c.tls_session,  flags);
 	if (ret != GNUTLS_E_SUCCESS) {
 		tls_client_ctx_free(ctx);
