@@ -6,7 +6,7 @@
 
 #include "daemon/engine.h"
 #include "lib/generic/array.h"
-#include "lib/generic/map.h"
+#include "lib/generic/trie.h"
 
 
 /** Query resolution task (opaque). */
@@ -17,6 +17,8 @@ struct worker_ctx;
 struct session;
 /** Zone import context (opaque). */
 struct zone_import_ctx;
+/** Data about the communication (defined in io.h). */
+struct io_comm_data;
 
 /** Pointer to the singleton worker.  NULL if not initialized. */
 KR_EXPORT extern struct worker_ctx *the_worker;
@@ -31,15 +33,14 @@ void worker_deinit(void);
 /**
  * Process an incoming packet (query from a client or answer from upstream).
  *
- * @param session  session the packet came from, or NULL (not from network)
- * @param peer     address the packet came from, or NULL (not from network)
- * @param eth_*    MAC addresses or NULL (they're useful for XDP)
- * @param pkt      the packet, or NULL (an error from the transport layer)
+ * @param session     session the packet came from, or NULL (not from network)
+ * @param comm        IO communication data (see `struct io_comm_data` docs)
+ * @param eth_*       MAC addresses or NULL (they're useful for XDP)
+ * @param pkt         the packet, or NULL (an error from the transport layer)
  * @return 0 or an error code
  */
-int worker_submit(struct session *session,
-		  const struct sockaddr *peer, const struct sockaddr *dst_addr,
-		  const uint8_t *eth_from, const uint8_t *eth_to, knot_pkt_t *pkt);
+int worker_submit(struct session *session, struct io_comm_data *comm,
+                  const uint8_t *eth_from, const uint8_t *eth_to, knot_pkt_t *pkt);
 
 /**
  * End current DNS/TCP session, this disassociates pending tasks from this session
@@ -176,13 +177,12 @@ struct worker_ctx {
 
 	struct worker_stats stats;
 
-	struct zone_import_ctx* z_import;
 	bool too_many_open;
 	size_t rconcurrent_highwatermark;
 	/** List of active outbound TCP sessions */
-	map_t tcp_connected;
+	trie_t *tcp_connected;
 	/** List of outbound TCP sessions waiting to be accepted */
-	map_t tcp_waiting;
+	trie_t *tcp_waiting;
 	/** Subrequest leaders (struct qr_task*), indexed by qname+qtype+qclass. */
 	trie_t *subreq_out;
 	mp_freelist_t pool_mp;
