@@ -43,8 +43,6 @@ const log_group_names_t log_group_names[] = {
 	GRP_NAME_ITEM(LOG_GRP_GNUTLS),
 	GRP_NAME_ITEM(LOG_GRP_TLSCLIENT),
 	GRP_NAME_ITEM(LOG_GRP_XDP),
-	GRP_NAME_ITEM(LOG_GRP_ZIMPORT),
-	GRP_NAME_ITEM(LOG_GRP_ZSCANNER),
 	GRP_NAME_ITEM(LOG_GRP_DOH),
 	GRP_NAME_ITEM(LOG_GRP_DNSSEC),
 	GRP_NAME_ITEM(LOG_GRP_HINT),
@@ -79,6 +77,7 @@ const log_group_names_t log_group_names[] = {
 	GRP_NAME_ITEM(LOG_GRP_MODULE),
 	GRP_NAME_ITEM(LOG_GRP_DEVEL),
 	GRP_NAME_ITEM(LOG_GRP_RENUMBER),
+	GRP_NAME_ITEM(LOG_GRP_EDE),
 	GRP_NAME_ITEM(LOG_GRP_REQDBG),
 	{ NULL, LOG_GRP_UNKNOWN },
 };
