@@ -11,7 +11,6 @@
 #include <libknot/descriptor.h>
 #include <ucw/mempool.h>
 #include <sys/socket.h>
-#include "kresconfig.h"
 #include "lib/resolve.h"
 #include "lib/layer.h"
 #include "lib/rplan.h"
@@ -26,7 +25,7 @@
 #define KNOT_EDNS_OPTION_COOKIE 10
 #endif /* ENABLE_COOKIES */
 
-#define VERBOSE_MSG(qry, ...) QRVERBOSE((qry), RESOLVER,  __VA_ARGS__)
+#define VERBOSE_MSG(qry, ...) kr_log_q((qry), RESOLVER,  __VA_ARGS__)
 
 bool kr_rank_check(uint8_t rank)
 {
@@ -346,6 +345,19 @@ static int edns_erase_and_reserve(knot_pkt_t *pkt)
 	return knot_pkt_reserve(pkt, len);
 }
 
+static inline size_t edns_padding_option_size(int32_t tls_padding)
+{
+	if (tls_padding == -1)
+		/* FIXME: we do not know how to reserve space for the
+		 * default padding policy, since we can't predict what
+		 * it will select. So i'm just guessing :/ */
+		return KNOT_EDNS_OPTION_HDRLEN + 512;
+	if (tls_padding >= 2)
+		return KNOT_EDNS_OPTION_HDRLEN + tls_padding;
+
+	return 0;
+}
+
 static int edns_create(knot_pkt_t *pkt, const struct kr_request *req)
 {
 	pkt->opt_rr = knot_rrset_copy(req->ctx->upstream_opt_rr, &pkt->mm);
@@ -356,14 +368,8 @@ static int edns_create(knot_pkt_t *pkt, const struct kr_request *req)
 		wire_size += KR_COOKIE_OPT_MAX_LEN;
 	}
 #endif /* ENABLE_COOKIES */
-	if (req->qsource.flags.tls) {
-		if (req->ctx->tls_padding == -1)
-			/* FIXME: we do not know how to reserve space for the
-			 * default padding policy, since we can't predict what
-			 * it will select. So i'm just guessing :/ */
-			wire_size += KNOT_EDNS_OPTION_HDRLEN + 512;
-		if (req->ctx->tls_padding >= 2)
-			wire_size += KNOT_EDNS_OPTION_HDRLEN + req->ctx->tls_padding;
+	if (req->qsource.flags.tls || req->qsource.comm_flags.tls) {
+		wire_size += edns_padding_option_size(req->ctx->tls_padding);
 	}
 	return knot_pkt_reserve(pkt, wire_size);
 }
@@ -417,26 +423,17 @@ static int write_extra_ranked_records(const ranked_rr_array_t *arr, uint16_t reo
 	return err;
 }
 
-/** @internal Add an EDNS padding RR into the answer if requested and required. */
-static int answer_padding(struct kr_request *request)
+static int pkt_padding(knot_pkt_t *packet, int32_t padding)
 {
-	if (kr_fails_assert(request && request->answer && request->ctx))
-		return kr_error(EINVAL);
-	if (!request->qsource.flags.tls) {
-		/* Not meaningful to pad without encryption. */
-		return kr_ok();
-	}
-	int32_t padding = request->ctx->tls_padding;
-	knot_pkt_t *answer = request->answer;
-	knot_rrset_t *opt_rr = answer->opt_rr;
+	knot_rrset_t *opt_rr = packet->opt_rr;
 	int32_t pad_bytes = -1;
 
 	if (padding == -1) { /* use the default padding policy from libknot */
-		pad_bytes =  knot_pkt_default_padding_size(answer, opt_rr);
+		pad_bytes =  knot_pkt_default_padding_size(packet, opt_rr);
 	}
 	if (padding >= 2) {
-		int32_t max_pad_bytes = knot_edns_get_payload(opt_rr) - (answer->size + knot_rrset_size(opt_rr));
-		pad_bytes = MIN(knot_edns_alignment_size(answer->size, knot_rrset_size(opt_rr), padding),
+		int32_t max_pad_bytes = knot_edns_get_payload(opt_rr) - (packet->size + knot_rrset_size(opt_rr));
+		pad_bytes = MIN(knot_edns_alignment_size(packet->size, knot_rrset_size(opt_rr), padding),
 				max_pad_bytes);
 	}
 
@@ -444,15 +441,27 @@ static int answer_padding(struct kr_request *request)
 		uint8_t zeros[MAX(1, pad_bytes)];
 		memset(zeros, 0, sizeof(zeros));
 		int r = knot_edns_add_option(opt_rr, KNOT_EDNS_OPTION_PADDING,
-					     pad_bytes, zeros, &answer->mm);
+					     pad_bytes, zeros, &packet->mm);
 		if (r != KNOT_EOK) {
-			knot_rrset_clear(opt_rr, &answer->mm);
+			knot_rrset_clear(opt_rr, &packet->mm);
 			return kr_error(r);
 		}
 	}
 	return kr_ok();
 }
 
+/** @internal Add an EDNS padding RR into the answer if requested and required. */
+static int answer_padding(struct kr_request *request)
+{
+	if (kr_fails_assert(request && request->answer && request->ctx))
+		return kr_error(EINVAL);
+	if (!request->qsource.flags.tls && !request->qsource.comm_flags.tls) {
+		/* Not meaningful to pad without encryption. */
+		return kr_ok();
+	}
+	return pkt_padding(request->answer, request->ctx->tls_padding);
+}
+
 /* Make a clean SERVFAIL answer. */
 static void answer_fail(struct kr_request *request)
 {
@@ -538,27 +547,26 @@ static void answer_finalize(struct kr_request *request)
 	/* AD flag.  We can only change `secure` from true to false.
 	 * Be conservative.  Primary approach: check ranks of all RRs in wire.
 	 * Only "negative answers" need special handling. */
-	bool secure = last != NULL && request->state == KR_STATE_DONE /*< suspicious otherwise */
+	bool secure = request->state == KR_STATE_DONE /*< suspicious otherwise */
 		&& knot_pkt_qtype(answer) != KNOT_RRTYPE_RRSIG;
-	if (last && (last->flags.STUB)) {
+	if (last->flags.STUB) {
 		secure = false; /* don't trust forwarding for now */
 	}
-	if (last && (last->flags.DNSSEC_OPTOUT)) {
+	if (last->flags.DNSSEC_OPTOUT) {
 		VERBOSE_MSG(last, "insecure because of opt-out\n");
 		secure = false; /* the last answer is insecure due to opt-out */
 	}
 
 	/* Write all RRsets meant for the answer. */
-	const uint16_t reorder = last ? last->reorder : 0;
 	bool answ_all_cnames = false/*arbitrary*/;
 	if (knot_pkt_begin(answer, KNOT_ANSWER)
-	    || write_extra_ranked_records(&request->answ_selected, reorder,
+	    || write_extra_ranked_records(&request->answ_selected, last->reorder,
 					answer, &secure, &answ_all_cnames)
 	    || knot_pkt_begin(answer, KNOT_AUTHORITY)
-	    || write_extra_ranked_records(&request->auth_selected, reorder,
+	    || write_extra_ranked_records(&request->auth_selected, last->reorder,
 					answer, &secure, NULL)
 	    || knot_pkt_begin(answer, KNOT_ADDITIONAL)
-	    || write_extra_ranked_records(&request->add_selected, reorder,
+	    || write_extra_ranked_records(&request->add_selected, last->reorder,
 					answer, NULL/*not relevant to AD*/, NULL)
 	    || answer_append_edns(request)
 	   )
@@ -567,7 +575,6 @@ static void answer_finalize(struct kr_request *request)
 		return;
 	}
 
-	if (!last) secure = false; /*< should be no-op, mostly documentation */
 	/* AD: "negative answers" need more handling. */
 	if (kr_response_classify(answer) != PKT_NOERROR
 	    /* Additionally check for CNAME chains that "end in NODATA",
@@ -722,6 +729,10 @@ knot_rrset_t* kr_request_ensure_edns(struct kr_request *request)
 
 knot_pkt_t *kr_request_ensure_answer(struct kr_request *request)
 {
+	if (request->options.NO_ANSWER) {
+		kr_assert(request->state & KR_STATE_FAIL);
+		return NULL;
+	}
 	if (request->answer)
 		return request->answer;
 
@@ -731,9 +742,10 @@ knot_pkt_t *kr_request_ensure_answer(struct kr_request *request)
 	// Find answer_max: limit on DNS wire length.
 	uint16_t answer_max;
 	const struct kr_request_qsource_flags *qs_flags = &request->qsource.flags;
-	if (kr_fails_assert((qs_flags->tls || qs_flags->http) ? qs_flags->tcp : true))
+	const struct kr_request_qsource_flags *qs_cflags = &request->qsource.comm_flags;
+	if (kr_fails_assert(!(qs_flags->tls || qs_cflags->tls || qs_cflags->http) || qs_flags->tcp))
 		goto fail;
-	if (!request->qsource.addr || qs_flags->tcp) {
+	if (!request->qsource.addr || qs_flags->tcp || qs_cflags->tcp) {
 		// not on UDP
 		answer_max = KNOT_WIRE_MAX_PKTSIZE;
 	} else if (knot_pkt_has_edns(qs_pkt)) {
@@ -780,16 +792,6 @@ fail:
 	return request->answer = NULL;
 }
 
-static bool resolution_time_exceeded(struct kr_query *qry, uint64_t now)
-{
-	uint64_t resolving_time = now - qry->creation_time_mono;
-	if (resolving_time > KR_RESOLVE_TIME_LIMIT) {
-		VERBOSE_MSG(qry, "query resolution time limit exceeded\n");
-		return true;
-	}
-	return false;
-}
-
 int kr_resolve_consume(struct kr_request *request, struct kr_transport **transport, knot_pkt_t *packet)
 {
 	struct kr_rplan *rplan = &request->rplan;
@@ -802,36 +804,36 @@ int kr_resolve_consume(struct kr_request *request, struct kr_transport **transpo
 	/* Different processing for network error */
 	struct kr_query *qry = array_tail(rplan->pending);
 	/* Check overall resolution time */
-	if (resolution_time_exceeded(qry, kr_now())) {
+	if (kr_now() - qry->creation_time_mono >= KR_RESOLVE_TIME_LIMIT) {
+		kr_query_inform_timeout(request, qry);
 		return KR_STATE_FAIL;
 	}
 	bool tried_tcp = (qry->flags.TCP);
-	if (!packet || packet->size == 0) {
+	if (!packet || packet->size == 0)
 		return KR_STATE_PRODUCE;
+
+	/* Packet cleared, derandomize QNAME. */
+	knot_dname_t *qname_raw = knot_pkt_qname(packet);
+	if (qname_raw && qry->secret != 0) {
+		randomized_qname_case(qname_raw, qry->secret);
+	}
+	request->state = KR_STATE_CONSUME;
+	if (qry->flags.CACHED) {
+		ITERATE_LAYERS(request, qry, consume, packet);
 	} else {
-		/* Packet cleared, derandomize QNAME. */
-		knot_dname_t *qname_raw = knot_pkt_qname(packet);
-		if (qname_raw && qry->secret != 0) {
-			randomized_qname_case(qname_raw, qry->secret);
-		}
-		request->state = KR_STATE_CONSUME;
-		if (qry->flags.CACHED) {
-			ITERATE_LAYERS(request, qry, consume, packet);
-		} else {
-			/* Fill in source and latency information. */
-			request->upstream.rtt = kr_now() - qry->timestamp_mono;
-			request->upstream.transport = transport ? *transport : NULL;
-			ITERATE_LAYERS(request, qry, consume, packet);
-			/* Clear temporary information */
-			request->upstream.transport = NULL;
-			request->upstream.rtt = 0;
-		}
+		/* Fill in source and latency information. */
+		request->upstream.rtt = kr_now() - qry->timestamp_mono;
+		request->upstream.transport = transport ? *transport : NULL;
+		ITERATE_LAYERS(request, qry, consume, packet);
+		/* Clear temporary information */
+		request->upstream.transport = NULL;
+		request->upstream.rtt = 0;
 	}
 
 	if (transport && !qry->flags.CACHED) {
 		if (!(request->state & KR_STATE_FAIL)) {
 			/* Do not complete NS address resolution on soft-fail. */
-			const int rcode = packet ? knot_wire_get_rcode(packet->wire) : 0;
+			const int rcode = knot_wire_get_rcode(packet->wire);
 			if (rcode != KNOT_RCODE_SERVFAIL && rcode != KNOT_RCODE_REFUSED) {
 				qry->flags.AWAIT_IPV6 = false;
 				qry->flags.AWAIT_IPV4 = false;
@@ -915,8 +917,8 @@ static struct kr_query *zone_cut_subreq(struct kr_rplan *rplan, struct kr_query
 static int forward_trust_chain_check(struct kr_request *request, struct kr_query *qry, bool resume)
 {
 	struct kr_rplan *rplan = &request->rplan;
-	map_t *trust_anchors = &request->ctx->trust_anchors;
-	map_t *negative_anchors = &request->ctx->negative_anchors;
+	trie_t *trust_anchors = request->ctx->trust_anchors;
+	trie_t *negative_anchors = request->ctx->negative_anchors;
 
 	if (qry->parent != NULL &&
 	    !(qry->forward_flags.CNAME) &&
@@ -1101,8 +1103,8 @@ static int forward_trust_chain_check(struct kr_request *request, struct kr_query
 static int trust_chain_check(struct kr_request *request, struct kr_query *qry)
 {
 	struct kr_rplan *rplan = &request->rplan;
-	map_t *trust_anchors = &request->ctx->trust_anchors;
-	map_t *negative_anchors = &request->ctx->negative_anchors;
+	trie_t *trust_anchors = request->ctx->trust_anchors;
+	trie_t *negative_anchors = request->ctx->negative_anchors;
 
 	/* Disable DNSSEC if it enters NTA. */
 	if (kr_ta_get(negative_anchors, qry->zone_cut.name)){
@@ -1534,6 +1536,17 @@ int kr_resolve_checkout(struct kr_request *request, const struct sockaddr *src,
 
 	/* Write down OPT unless in safemode */
 	if (!(qry->flags.NO_EDNS)) {
+		/* TLS padding */
+		if (transport->protocol == KR_TRANSPORT_TLS) {
+			size_t padding_size = edns_padding_option_size(request->ctx->tls_padding);
+			ret = knot_pkt_reserve(packet, padding_size);
+			if (ret)
+				return kr_error(EINVAL);
+			ret = pkt_padding(packet, request->ctx->tls_padding);
+			if (ret)
+				return kr_error(EINVAL);
+		}
+
 		ret = edns_put(packet, true);
 		if (ret != 0) {
 			return kr_error(EINVAL);
@@ -1615,4 +1628,71 @@ knot_mm_t *kr_resolve_pool(struct kr_request *request)
 	return NULL;
 }
 
+static int ede_priority(int info_code)
+{
+	switch(info_code) {
+	case KNOT_EDNS_EDE_DNSKEY_BIT:
+	case KNOT_EDNS_EDE_DNSKEY_MISS:
+	case KNOT_EDNS_EDE_SIG_EXPIRED:
+	case KNOT_EDNS_EDE_SIG_NOTYET:
+	case KNOT_EDNS_EDE_RRSIG_MISS:
+	case KNOT_EDNS_EDE_NSEC_MISS:
+		return 900;  /* Specific DNSSEC failures */
+	case KNOT_EDNS_EDE_BOGUS:
+		return 800;  /* Generic DNSSEC failure */
+	case KNOT_EDNS_EDE_FORGED:
+	case KNOT_EDNS_EDE_FILTERED:
+		return 700;  /* Considered hard fail by firefox */
+	case KNOT_EDNS_EDE_PROHIBITED:
+	case KNOT_EDNS_EDE_BLOCKED:
+	case KNOT_EDNS_EDE_CENSORED:
+		return 600;  /* Policy related */
+	case KNOT_EDNS_EDE_DNSKEY_ALG:
+	case KNOT_EDNS_EDE_DS_DIGEST:
+		return 500;  /* Non-critical DNSSEC issues */
+	case KNOT_EDNS_EDE_STALE:
+	case KNOT_EDNS_EDE_STALE_NXD:
+		return 300;  /* Serve-stale answers. */
+	case KNOT_EDNS_EDE_INDETERMINATE:
+	case KNOT_EDNS_EDE_CACHED_ERR:
+	case KNOT_EDNS_EDE_NOT_READY:
+	case KNOT_EDNS_EDE_NOTAUTH:
+	case KNOT_EDNS_EDE_NOTSUP:
+	case KNOT_EDNS_EDE_NREACH_AUTH:
+	case KNOT_EDNS_EDE_NETWORK:
+	case KNOT_EDNS_EDE_INV_DATA:
+		return 200;  /* Assorted codes */
+	case KNOT_EDNS_EDE_OTHER:
+		return 100;  /* Most generic catch-all error */
+	case KNOT_EDNS_EDE_NONE:
+		return 0;  /* No error - allow overriding */
+	default:
+		kr_assert(false);  /* Unknown info_code */
+		return 50;
+	}
+}
+
+int kr_request_set_extended_error(struct kr_request *request, int info_code, const char *extra_text)
+{
+	if (kr_fails_assert(request))
+		return KNOT_EDNS_EDE_NONE;
+
+	struct kr_extended_error *ede = &request->extended_error;
+
+	/* Clear any previously set error. */
+	if (info_code == KNOT_EDNS_EDE_NONE) {
+		kr_assert(extra_text == NULL);
+		ede->info_code = KNOT_EDNS_EDE_NONE;
+		ede->extra_text = NULL;
+		return KNOT_EDNS_EDE_NONE;
+	}
+
+	if (ede_priority(info_code) >= ede_priority(ede->info_code)) {
+		ede->info_code = info_code;
+		ede->extra_text = extra_text;
+	}
+
+	return ede->info_code;
+}
+
 #undef VERBOSE_MSG
