@@ -3,23 +3,18 @@
 #include "lib/selection.h"
 #include "lib/selection_forward.h"
 #include "lib/selection_iter.h"
-#include "lib/generic/pack.h"
-#include "lib/generic/trie.h"
 #include "lib/rplan.h"
 #include "lib/cache/api.h"
 #include "lib/resolve.h"
 
-#include "daemon/worker.h"
-#include "daemon/tls.h"
-
 #include "lib/utils.h"
 
-#define VERBOSE_MSG(qry, ...) QRVERBOSE((qry), SELECTION, __VA_ARGS__)
+#define VERBOSE_MSG(qry, ...) kr_log_q((qry), SELECTION, __VA_ARGS__)
 
 #define DEFAULT_TIMEOUT 400
 #define MAX_TIMEOUT 10000
 #define EXPLORE_TIMEOUT_COEFFICIENT 2
-#define MAX_BACKOFF 5
+#define MAX_BACKOFF 8
 #define MINIMAL_TIMEOUT_ADDITION 20
 
 /* After TCP_TIMEOUT_THRESHOLD timeouts one transport, we'll switch to TCP. */
@@ -81,7 +76,7 @@ static struct {
 	uint8_t addr_prefixes[NO6_PREFIX_COUNT][NO6_PREFIX_BYTES];
 } no6_est = { .len_used = 0 };
 
-static inline bool no6_is_bad(void)
+bool no6_is_bad(void)
 {
 	return no6_est.len_used == NO6_PREFIX_COUNT;
 }
@@ -180,7 +175,7 @@ int put_rtt_state(const uint8_t *ip, size_t len, struct rtt_state state,
 	return ret;
 }
 
-void bytes_to_ip(uint8_t *bytes, size_t len, uint16_t port, union inaddr *dst)
+void bytes_to_ip(uint8_t *bytes, size_t len, uint16_t port, union kr_sockaddr *dst)
 {
 	switch (len) {
 	case sizeof(struct in_addr):
@@ -199,7 +194,7 @@ void bytes_to_ip(uint8_t *bytes, size_t len, uint16_t port, union inaddr *dst)
 	}
 }
 
-uint8_t *ip_to_bytes(const union inaddr *src, size_t len)
+uint8_t *ip_to_bytes(const union kr_sockaddr *src, size_t len)
 {
 	switch (len) {
 	case sizeof(struct in_addr):
@@ -254,7 +249,7 @@ static void invalidate_dead_upstream(struct address_state *state,
 				     unsigned int retry_timeout)
 {
 	struct rtt_state *rs = &state->rtt_state;
-	if (rs->consecutive_timeouts >= KR_NS_TIMEOUT_ROW_DEAD) {
+	if (rs->dead_since) {
 		uint64_t now = kr_now();
 		if (now < rs->dead_since) {
 			// broken continuity of timestamp (reboot, different machine, etc.)
@@ -311,7 +306,7 @@ static void check_network_settings(struct address_state *address_state,
 	}
 }
 
-void update_address_state(struct address_state *state, union inaddr *address,
+void update_address_state(struct address_state *state, union kr_sockaddr *address,
 			  size_t address_len, struct kr_query *qry)
 {
 	check_tls_capable(state, qry->request, &address->ip);
@@ -335,11 +330,8 @@ void update_address_state(struct address_state *state, union inaddr *address,
 #endif
 }
 
-static int cmp_choices(const void *a, const void *b)
+static int cmp_choices(const struct choice *a_, const struct choice *b_)
 {
-	const struct choice *a_ = a;
-	const struct choice *b_ = b;
-
 	int diff;
 	/* Prefer IPv4 if IPv6 appears to be generally broken. */
 	diff = (int)a_->address_len - (int)b_->address_len;
@@ -364,21 +356,40 @@ static int cmp_choices(const void *a, const void *b)
 	}
 	return 0;
 }
-
-/* Fisher-Yates shuffle of the choices */
-static void shuffle_choices(struct choice choices[], int choices_len)
+/** Select the best entry from choices[] according to cmp_choices() comparator.
+ *
+ * Ties are decided in an (almost) uniformly random fashion.
+ */
+static const struct choice * select_best(const struct choice choices[], int choices_len)
 {
-	struct choice tmp;
-	for (int i = choices_len - 1; i > 0; i--) {
-		int j = kr_rand_bytes(1) % (i + 1);
-		tmp = choices[i];
-		choices[i] = choices[j];
-		choices[j] = tmp;
+	/* Deciding ties: it's as-if each index carries one byte of randomness.
+	 * Ties get decided by comparing that byte, and the byte itself
+	 * is computed lazily (negative until computed).
+	 */
+	int best_i = 0;
+	int best_rnd = -1;
+	for (int i = 1; i < choices_len; ++i) {
+		int diff = cmp_choices(&choices[i], &choices[best_i]);
+		if (diff > 0)
+			continue;
+		if (diff < 0) {
+			best_i = i;
+			best_rnd = -1;
+			continue;
+		}
+		if (best_rnd < 0)
+			best_rnd = kr_rand_bytes(1);
+		int new_rnd = kr_rand_bytes(1);
+		if (new_rnd < best_rnd) {
+			best_i = i;
+			best_rnd = new_rnd;
+		}
 	}
+	return &choices[best_i];
 }
 
 /* Adjust choice from `unresolved` in case of NO6 (broken IPv6). */
-static struct kr_transport unresolved_adjust(struct to_resolve unresolved[],
+static struct kr_transport unresolved_adjust(const struct to_resolve unresolved[],
 					     int unresolved_len, int index)
 {
 	if (unresolved[index].type != KR_TRANSPORT_RESOLVE_AAAA || !no6_is_bad())
@@ -409,8 +420,8 @@ finish:
 }
 
 /* Performs the actual selection (currently variation on epsilon-greedy). */
-struct kr_transport *select_transport(struct choice choices[], int choices_len,
-				      struct to_resolve unresolved[],
+struct kr_transport *select_transport(const struct choice choices[], int choices_len,
+				      const struct to_resolve unresolved[],
 				      int unresolved_len, int timeouts,
 				      struct knot_mm *mempool, bool tcp,
 				      size_t *choice_index)
@@ -422,17 +433,16 @@ struct kr_transport *select_transport(struct choice choices[], int choices_len,
 
 	struct kr_transport *transport = mm_calloc(mempool, 1, sizeof(*transport));
 
-	/* Shuffle, so we choose fairly between choices with same attributes. */
-	shuffle_choices(choices, choices_len);
 	/* If there are some addresses with no rtt_info we try them
 	 * first (see cmp_choices). So unknown servers are chosen
 	 * *before* the best know server. This ensures that every option
 	 * is tried before going back to some that was tried before. */
-	qsort(choices, choices_len, sizeof(struct choice), cmp_choices);
-	struct choice *best = &choices[0];
-	struct choice *chosen;
+	const struct choice *best = select_best(choices, choices_len);
+	const struct choice *chosen;
 
-	const bool explore = choices_len == 0 || kr_rand_coin(EPSILON_NOMIN, EPSILON_DENOM);
+	const bool explore = choices_len == 0 || kr_rand_coin(EPSILON_NOMIN, EPSILON_DENOM)
+		/* We may need to explore to get at least one A record. */
+		|| (no6_is_bad() && best->address.ip.sa_family == AF_INET6);
 	if (explore) {
 		/* "EXPLORE":
 		 * randomly choose some option
@@ -449,8 +459,6 @@ struct kr_transport *select_transport(struct choice choices[], int choices_len,
 		/* "EXPLOIT":
 		 * choose a resolved address which seems best right now. */
 		chosen = best;
-		if (no6_is_bad())
-			VERBOSE_MSG(NULL, "NO6: is KO [exploit]\n");
 	}
 
 	/* Don't try the same server again when there are other choices to be explored */
@@ -576,35 +584,49 @@ void update_rtt(struct kr_query *qry, struct address_state *addr_state,
 	}
 }
 
-static void cache_timeout(const struct kr_query *qry, const struct kr_transport *transport,
+/// Update rtt_state (including caching) after a server timed out.
+static void server_timeout(const struct kr_query *qry, const struct kr_transport *transport,
 			  struct address_state *addr_state, struct kr_cache *cache)
 {
-	if (transport->deduplicated) {
-		/* Transport was chosen by a different query, that one will
-		 * cache the result. */
+	// Make sure that the timeout wasn't capped; see kr_transport::timeout_capped
+	if (transport->timeout_capped)
 		return;
-	}
 
-	uint8_t *address = ip_to_bytes(&transport->address, transport->address_len);
+	const uint8_t *address = ip_to_bytes(&transport->address, transport->address_len);
 	if (transport->address_len == sizeof(struct in6_addr))
 		no6_timed_out(qry, address);
 
-	struct rtt_state old_state = addr_state->rtt_state;
-	struct rtt_state cur_state =
-		get_rtt_state(address, transport->address_len, cache);
+	struct rtt_state *state = &addr_state->rtt_state;
+	// While we were waiting for timeout, the stats might have changed considerably,
+	// so let's overwrite what we had by fresh cache contents.
+	// This is useful when the address is busy (we query it concurrently).
+	*state = get_rtt_state(address, transport->address_len, cache);
 
-	/* We could lose some update from some other process by doing this,
-	 * but at least timeout count can't blow up. */
-	if (cur_state.consecutive_timeouts == old_state.consecutive_timeouts) {
-		if (++cur_state.consecutive_timeouts >=
-		    KR_NS_TIMEOUT_ROW_DEAD) {
-			cur_state.dead_since = kr_now();
-		}
-		put_rtt_state(address, transport->address_len, cur_state, cache);
-	} else {
-		/* `get_rtt_state` opens a cache transaction, we have to end it. */
-		kr_cache_commit(cache);
+	++state->consecutive_timeouts;
+	// Avoid overflow; we don't utilize very high values anyway (arbitrary limit).
+	state->consecutive_timeouts = MIN(64, state->consecutive_timeouts);
+	if (state->consecutive_timeouts >= KR_NS_TIMEOUT_ROW_DEAD) {
+		// Only mark as dead if we waited long enough,
+		// so that many (concurrent) short attempts can't cause the dead state.
+		if (transport->timeout >= KR_NS_TIMEOUT_MIN_DEAD_TIMEOUT)
+			state->dead_since = kr_now();
 	}
+
+	// If transport was chosen by a different query, that one will cache it.
+	if (!transport->deduplicated) {
+		put_rtt_state(address, transport->address_len, *state, cache);
+	} else {
+		kr_cache_commit(cache); // Avoid any risk of long transaction.
+	}
+}
+// Not everything can be checked in nice ways like static_assert()
+static __attribute__((constructor)) void test_RTT_consts(void)
+{
+	// See KR_NS_TIMEOUT_MIN_DEAD_TIMEOUT above.
+	kr_require(
+		calc_timeout((struct rtt_state){ .consecutive_timeouts = MAX_BACKOFF, })
+		 >= KR_NS_TIMEOUT_MIN_DEAD_TIMEOUT
+	);
 }
 
 void error(struct kr_query *qry, struct address_state *addr_state,
@@ -629,12 +651,7 @@ void error(struct kr_query *qry, struct address_state *addr_state,
 	case KR_SELECTION_TLS_HANDSHAKE_FAILED:
 	case KR_SELECTION_QUERY_TIMEOUT:
 		qry->server_selection.local_state->timeouts++;
-		/* Make sure that the query was chosen by this query and timeout wasn't capped
-		 * (see kr_transport::timeout_capped for details). */
-		if (!transport->deduplicated && !transport->timeout_capped) {
-			cache_timeout(qry, transport, addr_state,
-				      &qry->request->ctx->cache);
-		}
+		server_timeout(qry, transport, addr_state, &qry->request->ctx->cache);
 		break;
 	case KR_SELECTION_FORMERR:
 		if (qry->flags.NO_EDNS) {
@@ -751,7 +768,7 @@ int kr_forward_add_target(struct kr_request *req, const struct sockaddr *sock)
 		return kr_error(EINVAL);
 	}
 
-	union inaddr address;
+	union kr_sockaddr address;
 
 	switch (sock->sa_family) {
 	case AF_INET:
