@@ -5,12 +5,17 @@
 #include "lib/selection_forward.h"
 #include "lib/resolve.h"
 
-#define VERBOSE_MSG(qry, ...) QRVERBOSE((qry), SELECTION, __VA_ARGS__)
+#define VERBOSE_MSG(qry, ...) kr_log_q((qry), SELECTION, __VA_ARGS__)
 
 #define FORWARDING_TIMEOUT 2000
+/* TODO: well, this is a bit hard; maybe we'd better have a different approach
+ * for estimating DEAD-ness for forwarding.
+ * Even ACKs on connections might be useful here. */
+static_assert(FORWARDING_TIMEOUT >= KR_NS_TIMEOUT_MIN_DEAD_TIMEOUT,
+		"Bad combination of NS selection limits.");
 
 struct forward_local_state {
-	inaddr_array_t *targets;
+	kr_sockaddr_array_t *targets;
 	struct address_state *addr_states;
 	/** Index of last choice in the targets array, used for error reporting. */
 	size_t last_choice_index;
@@ -38,7 +43,7 @@ void forward_choose_transport(struct kr_query *qry,
 	int valid = 0;
 
 	for (int i = 0; i < local_state->targets->len; i++) {
-		union inaddr *address = &local_state->targets->at[i];
+		union kr_sockaddr *address = &local_state->targets->at[i];
 		size_t addr_len;
 		uint16_t port;
 		switch (address->ip.sa_family) {
