@@ -9,7 +9,7 @@ M.timeout = 3*sec
 
 M.callback = ffi.cast("kr_stale_cb",
 	function (ttl) --, name, type, qry)
-		--log('[     ][stal]   => called back with TTL: ' .. tostring(ttl))
+		--log_debug(ffi.C.SRVSTALE, '   => called back with TTL: ' .. tostring(ttl))
 		if ttl + 3600 * 24 > 0 then -- at most one day stale
 			return 1
 		else
@@ -27,9 +27,7 @@ M.layer = {
 		local now = ffi.C.kr_now()
 		local deadline = qry.creation_time_mono + M.timeout
 		if now > deadline or qry.flags.NO_NS_FOUND then
-			if verbose() then
-				log('[     ][stal]   => no reachable NS, using stale data')
-			end
+			log_debug(ffi.C.LOG_GRP_SRVSTALE, '   => no reachable NS, using stale data')
 			qry.stale_cb = M.callback
 			-- TODO: probably start the same request that doesn't stale-serve,
 			-- but first we need some detection of non-interactive / internal requests.
